express = require 'express'
zombie = require 'zombie'
http = require 'http'
send = require('connect').static.send
{parse} = require 'url'
fs = require 'fs'
{mkdirs} = require './mkdirs.js'
path = require 'path'
querystring = require 'querystring'

feral = {}

feral.create = (options) ->
  if options.listen? and typeof options.listen is 'function'
    server = options
    options = arguments[1]
    extend_server server, options
  else
    throw "port number required to create a Feral server" if not options.port 
    throw "public path required to create a Feral server" if not options.public
    server = create_server()
    extend_server server, options
    server.use express.static options.public
    server.use server.router
    server.listen options.port
    server.port = options.port
  console.log "Feral Express server listening on port #{options.port}"
  server

extend_server = (server,options) ->
  for method of InstanceMethods
    server[method] = InstanceMethods[method]
  server._feralTargets = {}
  for target_name, test of options.targets || {}
    server.createTarget target_name, test
  throw "cacheDir key required to create a Feral server" if not options.cacheDir
  server._feralCache = options.cacheDir
  server.use (request,response,next) ->
    cache_middleware.apply server, [request,response,next]
  throw "cache key required to create a Feral server" if not options.cache  
  for cache_name, cache_options of options.cache
    server.createCache cache_name, cache_options
  server

create_server = ->
  server = express.createServer()
  server.use express.methodOverride()
  server.use express.bodyParser()
  server.use express.cookieParser()
  server.use express.logger()
  server

cache_middleware = (request,response,next) ->
  options =
    root: @_feralCache
  for target_name, test of @_feralTargets
    if test request
      starting_slash = request.url.match(/^\//)
      if starting_slash
        options.path = "/#{target_name}#{request.url}"
      else
        options.path = "#{target_name}/#{request.url}"
      console.log "Feral serving cached file: #{options.path}"
      return send request, response, next, options
  next()
  
InstanceMethods = 
  createTarget: (name,condition) ->
    @_feralTargets[name] = condition
  createCache: (cache_name,cache_options) ->
     throw "url key required" if not cache_options.url?
     for method in ['get','post']
       do (method) =>
         @[method] "/feral/#{cache_name}", (request,response) =>
           generated_url = generate_url.call @, cache_options.url, request.query
           serialize_page request, generated_url, cache_options.callback, (browser) =>
             for target_name, target_callback of cache_options.targets || {}
               do (target_name, target_callback) =>
                 if typeof target_callback is 'function'
                   #TODO: fix: browser.fork() failing
                   forked_browser = browser
                   target_callback forked_browser, =>
                     write_cache.call @, target_name, parse(generated_url).pathname, forked_browser, response
                 else if target_callback is true
                   write_cache.call @, target_name, parse(generated_url).pathname, browser, response
  cache: (options) ->
    for cache_name, query_params of options
      http.get
        host: 'localhost'
        port: @port
        path: "/feral/#{cache_name}?#{querystring.stringify query_params || {}}"

#generate_url("/:slug/:page?",{slug:"tasty"}) == "/tasty"
generate_url = (url,params) ->
  url = url.replace(/\*/,params.path.replace(/^\//,'')) if params.path
  param_matcher = new RegExp('(\\()?\\:([\\w]+)(\\))?(\\?)?(/|$)','g')
  for param_name of params
    url = url.replace param_matcher, ->
      if arguments[2] is param_name
        params[param_name] + arguments[5]
      else
        (arguments[1] || '') + ':' + arguments[2] + (arguments[3] || '') + (arguments[4] || '') + arguments[5]
  #remove :key?
  optional_param_matcher = new RegExp('(\\()?\\:([\\w]+)(\\))?\\?(/|$)','g')
  url = url.replace optional_param_matcher, ''
  url.replace(/\([^\)]+\)/g,'')
  if !url.match /^https?:\/\//
    port = if @port? then ":#{@port}" else ""
    url = "http://localhost#{port}" + (if url.match(/^\//) then "" else "/") + url
  url

serialize_page = (request,url,callback,next) ->
  browser = new zombie.Browser
  browser.removeScript = (src) ->
    browser.css('head script' + if src then '[href="' + src + '"]' else '').map (script) ->
      script.parentNode.removeChild script
  browser.addScript = (src) ->
    script = browser.window.document.createElement 'script'
    script.type = 'text/javascript'
    script.src = src
    browser.css('head')[0].appendChild script
  browser.visit url, ->
    console.log "Feral / Zombie visited: #{url}"
    if callback
      callback browser, ->
        next browser
    else
      next browser

write_cache = (target_name, generated_url, browser, response) ->
  filename = @_feralCache + target_name + generated_url
  filename += 'index.html' if filename.match /\/$/
  mkdirs path.dirname(filename), (err,created) ->
    fs.writeFile filename, browser.html(), (err) ->
      response.send "Feral cached #{filename}"
  
http.IncomingMessage.prototype.agent = ->
  return @_agent if @_agent?
  
  name = 'unknown'
  version = "0"
  ua = @headers['user-agent'] || ''
  
  if ua.match(/Chrome/)
    name = 'chrome'
    version = ua.match(/Chrome\/([\d]+)/)
  else if ua.match(/Firefox/)
    name = 'firefox'
    version = ua.match(/Firefox\/([\d]+)/)
  else if ua.match(/Safari/)
    name = 'safari'
    version = ua.match(/Version\/([\d]+)/)
  else if ua.match(/(iPhone|iPad)/)
    name = 'ios'
    version = ua.match(/\sOS\s([\d]+)/)
  else if ua.match(/AppleWebKit/)
    name = 'safari'
  else if ua.match(/Android/)
    name = 'android'
    version = ua.match(/\s([\d\.]+)/)
  else if ua.match(/Opera/)
    name = 'opera'
    version = ua.match(/Version\/([\d\.]+)/)
  else if ua.match(/MSIE/)
    name = 'ie'
    version = ua.match(/MSIE\s([\d\.]+)/)
  else if ua.match(/bingbot/i)
    name = 'bing'
    version = ua.match(/bingbot\/([\d\.]+)/i)
  else if ua.match(/Googlebot/i)
    name = 'google'
    version = ua.match(/Googlebot\/([\d\.]+)/i)
  else if ua.match(/Ask Jeeves/i)
    name = 'ask'
  else if ua.match(/Yahoo/i)
    name = 'yahoo'

  version = version[1] if version
  bits = if version then version.split('.') else [0,0]
  
  history_available =
    (name is 'chrome' and parseInt(version) >= 8) or
    (name is 'safari' and parseInt(version) >= 5) or
    (name is 'firefox' and parseInt(version) >= 4) or
    (name is 'ios' and parseInt(version) >= 4) or
    (name is 'android' and (parseInt(bits[0]) > 2 or (parseInt(bits[0]) is 2 and parseInt(bits[1]) > 2))) or
    (name is 'opera' and (parseInt(bits[0]) > 11 or parseInt(bits[0]) is 11 and bits[1].match(/^[5-9]/)))
  
  @_agent = 
    name: name
    version: version
    targets: {}
    robot: name in ['bing','google','ask','yahoo']
    historyAvailable: history_available
  
  for target_name, test of @connection.server._feralTargets
    @_agent.targets[target_name] = test @

  @_agent
  
module.exports = feral